#define APU_PL1_VOL(x)		*((unsigned char*)0x4000)=(x);
#define APU_PL1_SWEEP(x)	*((unsigned char*)0x4001)=(x);
#define APU_PL1_LO(x)		*((unsigned char*)0x4002)=(x);
#define APU_PL1_HI(x)		*((unsigned char*)0x4003)=(x);
#define APU_PL2_VOL(x)		*((unsigned char*)0x4004)=(x);
#define APU_PL2_SWEEP(x)	*((unsigned char*)0x4005)=(x);
#define APU_PL2_LO(x)		*((unsigned char*)0x4006)=(x);
#define APU_PL2_HI(x)		*((unsigned char*)0x4007)=(x);
#define APU_TRI_LINEAR(x)	*((unsigned char*)0x4008)=(x);
#define APU_TRI_LO(x)		*((unsigned char*)0x400a)=(x);
#define APU_TRI_HI(x)		*((unsigned char*)0x400b)=(x);
#define APU_NOISE_VOL(x)	*((unsigned char*)0x400c)=(x);
#define APU_NOISE_LO(x)		*((unsigned char*)0x400e)=(x);
#define APU_NOISE_HI(x)		*((unsigned char*)0x400f)=(x);
#define APU_DMC_FREQ(x)		*((unsigned char*)0x4010)=(x);
#define APU_DMC_RAW(x)		*((unsigned char*)0x4011)=(x);
#define APU_DMC_START(x)	*((unsigned char*)0x4012)=(x);
#define APU_DMC_LEN(x)		*((unsigned char*)0x4013)=(x);
#define APU_SND_CHN(x)		*((unsigned char*)0x4015)=(x);


unsigned char pulse1_vol;
unsigned char pulse1_duty;
unsigned char pulse1_lsb;
unsigned char pulse1_msb;
unsigned char pulse1_msb_prev;
unsigned int pulse1_period;

unsigned char pulse2_vol;
unsigned char pulse2_duty;
unsigned char pulse2_lsb;
unsigned char pulse2_msb;
unsigned char pulse2_msb_prev;
unsigned int pulse2_period;

unsigned int tri_period;
unsigned char tri_vol;

unsigned char noise_vol;
unsigned char noise_period;
unsigned char noise_mode;



void apu_init(void)
{
	APU_SND_CHN(0x0f);		//enable all channels, stop DMC
	APU_TRI_LINEAR(0x80);	//disable triangle length counter
	APU_NOISE_HI(0x00);		//load noise length
	APU_PL1_VOL(0x30);		//volumes to 0
	APU_PL2_VOL(0x30);
	APU_NOISE_VOL(0x30);
	APU_PL1_SWEEP(0x08);	//no sweep
	APU_PL2_SWEEP(0x08);
	
	pulse1_vol=0;
	pulse1_duty=0;
	pulse1_lsb=0;
	pulse1_msb=0;
	pulse1_msb_prev=0xff;
	pulse1_period=0;

	pulse2_vol=0;
	pulse2_duty=0;
	pulse2_lsb=0;
	pulse2_msb=0;
	pulse2_msb_prev=0xff;
	pulse2_period=0;
	
	tri_period=0;
	tri_vol=0;
	
	noise_vol=0;
	noise_period=0;
	noise_mode=0;
}



void apu_clear(void)
{
	pulse1_vol=0;
	pulse2_vol=0;
	noise_vol=0;
	tri_vol=0;
}



void apu_update(void)
{
	//pulse1
	
	pulse1_lsb=pulse1_period&0xff;
	pulse1_msb=(pulse1_period>>8)&0x07;
	
	APU_PL1_VOL(0x30|pulse1_duty|pulse1_vol);
	
	APU_PL1_LO(pulse1_lsb);
	
	if(pulse1_msb_prev!=pulse1_msb)
	{
		pulse1_msb_prev=pulse1_msb;
		APU_PL1_HI(pulse1_msb);
	}
	
	//pulse2
	
	pulse2_lsb=pulse2_period&0xff;
	pulse2_msb=(pulse2_period>>8)&0x07;
	
	APU_PL2_VOL(0x30|pulse2_duty|pulse2_vol);
	
	APU_PL2_LO(pulse2_lsb);
	
	if(pulse2_msb_prev!=pulse2_msb)
	{
		pulse2_msb_prev=pulse2_msb;
		APU_PL2_HI(pulse2_msb);
	}
	
	//tri
	
	if(tri_vol)
	{
		APU_TRI_LINEAR(0xff);
		APU_TRI_LO(tri_period&0xff);
		APU_TRI_HI(tri_period>>8);
	}
	else
	{
		APU_TRI_LINEAR(0x80);
	}
	
	//noise
	
	APU_NOISE_VOL(0x30|(noise_vol&15));
	APU_NOISE_LO(noise_period);
}
