enum {
	NOTE_C=0,
	NOTE_Ch,
	NOTE_D,
	NOTE_Dh,
	NOTE_E,
	NOTE_F,
	NOTE_Fh,
	NOTE_G,
	NOTE_Gh,
	NOTE_A,
	NOTE_Ah,
	NOTE_B
};



#define MAX_ARP_STEPS		6

unsigned char arp_list[MAX_ARP_STEPS];
unsigned char arp_list_len;		//also write pointer
unsigned char arp_list_ptr;		//play pointer

unsigned char arp_list_prev;
unsigned char arp_list_len_prev;

unsigned char any_note_active;

unsigned char base_octave;
unsigned char base_note;
unsigned int base_pitch;
unsigned int base_pitch_to;

const unsigned int* pitch_table;

unsigned char arp_speed;
unsigned char arp_speed_cnt;

struct env_struct env_vol1;
struct env_struct env_vol2;
struct env_struct env_pit1;
struct env_struct env_pit2;

unsigned int lead_period;
unsigned char lead_vol;
unsigned char lead_duty;
unsigned char lead_gain;

enum {
	MONOMODE_ARP_MONO=0,
	MONOMODE_ARP_LEGATO,
	MONOMODE_MONO,
	MONOMODE_PORTA,
	MONOMODE_POLY
};

unsigned char mono_mode;
signed int porta_speed;

unsigned char lead_mute_at_release;

signed char pitch_wheel;
signed char pitch_wheel_to;
signed char pitch_wheel_speed;

signed int poly_pitch1;
signed int poly_pitch2;

unsigned char poly_base_note1;
unsigned char poly_base_note2;

signed int lead_dual_offset;



void arp_list_clear(void)
{
	arp_list_len=0;
}



void __fastcall__ arp_list_add(unsigned char note)
{
	if((mono_mode!=MONOMODE_ARP_MONO)&&(mono_mode!=MONOMODE_ARP_LEGATO)&&(mono_mode!=MONOMODE_POLY)) arp_list_len=0;
	
	if(arp_list_len<MAX_ARP_STEPS)
	{
		arp_list[arp_list_len++]=note;
	}
}



void lead_env_start(void)
{
	any_note_active=TRUE;
		
	arp_list_ptr=0;			//restart arpeggio
	
	env_start(&env_vol1);	//restart volume and pitch envelope
	env_start(&env_pit1);
}



void lead_env_stop(unsigned char mute)
{			
	any_note_active=FALSE;

	if(mute)
	{
		env_stop(&env_vol1);
	}
}



void lead_volenv_set(const signed char* data)
{
	env_set_data(&env_vol1,data);
	env_set_data(&env_vol2,data);
	env_stop(&env_vol1);
	env_stop(&env_vol2);
}



void lead_pitenv_set(const signed char* data)
{
	env_set_data(&env_pit1,data);
	env_set_data(&env_pit2,data);
	env_stop(&env_pit1);
	env_stop(&env_pit2);
}



void lead_init(unsigned char ntsc)
{
	pitch_table=(ntsc?pitch_table_ntsc:pitch_table_pal);

	lead_period=0;
	lead_vol=0;
	lead_duty=2<<6;
	lead_gain=15;
	
	base_octave=2;
	base_note=0;
	base_pitch=base_octave*12+base_note*16;
	base_pitch_to=base_pitch;
	
	pitch_wheel=0;
	pitch_wheel_to=0;
	pitch_wheel_speed=4;
	
	any_note_active=FALSE;
	
	arp_list_len=0;
	arp_list_prev=0xff;
	arp_list_len_prev=0;
	arp_list_ptr=0;
	arp_speed_cnt=0;
	arp_speed=1;
	
	lead_mute_at_release=TRUE;
	
	memfill(&env_vol1,0,sizeof(env_vol1));
	memfill(&env_vol2,0,sizeof(env_vol2));
	memfill(&env_pit1,0,sizeof(env_pit1));
	memfill(&env_pit2,0,sizeof(env_pit2));
	
	lead_volenv_set(NULL);
	lead_pitenv_set(NULL);
	
	mono_mode=MONOMODE_ARP_MONO;
	porta_speed=4096;
	
	lead_dual_offset=0;
}



void lead_pitch_wheel_update(void)
{
	if(pitch_wheel<pitch_wheel_to)
	{
		pitch_wheel+=pitch_wheel_speed;
		if(pitch_wheel>pitch_wheel_to) pitch_wheel=pitch_wheel_to;
	}
	if(pitch_wheel>pitch_wheel_to)
	{
		pitch_wheel-=pitch_wheel_speed;
		if(pitch_wheel<pitch_wheel_to) pitch_wheel=pitch_wheel_to;
	}
}
	
	
	
void lead_pitch_minmax(signed int* pitch)
{
	if(*pitch<0) *pitch=0;
	if(*pitch>=8*12*16) *pitch=8*12*16-1;
}



void lead_update_mono(void)
{
	unsigned char first_note;
	signed int off;
	
	first_note=FALSE;
	
	if(!any_note_active)
	{
		if(arp_list_len>0)
		{
			lead_env_start();
			
			first_note=TRUE;	//reset portamento
		}
	}
	else
	{
		if(arp_list_len==0)
		{
			lead_env_stop(lead_mute_at_release);
		}
	}
	
	//process envelopes
	
	env_update(&env_vol1);
	env_update(&env_pit1);
		
	if(mono_mode==MONOMODE_ARP_MONO||mono_mode==MONOMODE_MONO)	//for mono modes restart envelope with each note change as well
	{
		if(arp_list_len>0)
		{
			if(arp_list_prev!=arp_list[0]||arp_list_len_prev!=arp_list_len)
			{
				lead_env_start();
				arp_list_prev=arp_list[0];
				arp_list_len_prev=arp_list_len;
			}
		}
		else
		{
			arp_list_prev=0xff;
			arp_list_len_prev=arp_list_len;
		}
	}
			
	//apply envelopes
	
	lead_vol=env_vol1.out;

	if(arp_list_len>0)
	{
		if(arp_list_ptr>=arp_list_len) arp_list_ptr=0;
	
		base_note=base_octave*12+arp_list[arp_list_ptr];
		
		++arp_speed_cnt;
		
		if(arp_speed_cnt>=arp_speed) arp_speed_cnt=0;
		
		if(arp_speed_cnt==0) ++arp_list_ptr;
	
		if(base_note>=8*12) base_note=8*12-1;
	
		base_pitch_to=base_note*16+env_pit1.out;
	
		if(mono_mode!=MONOMODE_PORTA||first_note)
		{
			base_pitch=base_pitch_to;
		}
		else
		{
			if(base_pitch<base_pitch_to)
			{
				base_pitch+=porta_speed;
				if(base_pitch>base_pitch_to) base_pitch=base_pitch_to;
			}

			if(base_pitch>base_pitch_to)
			{
				base_pitch-=porta_speed;
				if(base_pitch<base_pitch_to) base_pitch=base_pitch_to;
			}
		}
	}
	
	//process pitch wheel and apply resulting pitch
	
	lead_pitch_wheel_update();

	off=base_pitch+pitch_wheel;
	
	lead_pitch_minmax(&off);
		
	lead_period=pitch_table[off];
	
	//set APU output
	
	if(lead_duty<0xff)
	{
		pulse1_duty=lead_duty;
		pulse1_vol =lead_vol*lead_gain/15;
		pulse1_period=lead_period;
		
		if(lead_dual_offset)
		{
			off+=lead_dual_offset;
			
			lead_pitch_minmax(&off);
			
			pulse2_duty=lead_duty;
			pulse2_vol=lead_vol;	
			pulse2_period=pitch_table[off];
		}
	}
	else
	{
		if(lead_vol)
		{
			tri_vol=lead_vol;
			tri_period=lead_period;
		}
	}
	
	//send echo
	
	if(!lead_dual_offset||lead_duty==0xff)
	{
		echo_send(1,lead_vol*lead_gain/15,lead_duty,lead_period);
	}
	else
	{
		echo_send(0,pulse1_vol,pulse1_duty,pulse1_period);
		echo_send(1,pulse2_vol,pulse2_duty,pulse2_period);
	}
}



//separate logic for dual voice mode to avoid overcomplication of the monophonic part

void lead_update_poly(void)
{
	if(arp_list_len==0)
	{
		if(lead_mute_at_release)
		{
			env_stop(&env_vol1);
			env_stop(&env_vol2);
		}
		
		poly_base_note1=0xff;
		poly_base_note2=0xff;
	}
	else
	{
		if(poly_base_note1==0xff)
		{
			poly_base_note1=arp_list[0];
			env_start(&env_vol1);
			env_start(&env_pit1);
		}
			
		if(arp_list_len>1)
		{
			if(poly_base_note2==0xff)
			{
				if(poly_base_note1==arp_list[0]) poly_base_note2=arp_list[1]; else poly_base_note2=arp_list[0]; 
				env_start(&env_vol2);
				env_start(&env_pit2);
			}
		}
		else
		{
			env_stop(&env_vol2);
			env_stop(&env_pit2);
			poly_base_note2=0xff;
		}
	}
	
	//update envelopes
	
	env_update(&env_vol1);
	env_update(&env_vol2);
	env_update(&env_pit1);
	env_update(&env_pit2);
	
	//process pitch wheel and calculate pitch
	
	lead_pitch_wheel_update();

	poly_pitch1=(base_octave*12+poly_base_note1)*16+env_pit1.out+pitch_wheel;
	poly_pitch2=(base_octave*12+poly_base_note2)*16+env_pit2.out+pitch_wheel;
	
	lead_pitch_minmax(&poly_pitch1);
	lead_pitch_minmax(&poly_pitch2);
	
	//set APU output
	
	pulse1_vol=env_vol1.out*lead_gain/15;
	pulse2_vol=env_vol2.out*lead_gain/15;
	pulse1_duty=lead_duty<0xff?lead_duty:(2<<6);
	pulse2_duty=pulse1_duty;
	pulse1_period=pitch_table[poly_pitch1];
	pulse2_period=pitch_table[poly_pitch2];
	
	//send dual echo
	
	echo_send(0,pulse1_vol,pulse1_duty,pulse1_period);
	echo_send(1,pulse2_vol,pulse2_duty,pulse2_period);
}



void lead_update(void)
{
	if((mono_mode==MONOMODE_POLY)&&(lead_duty<0xff))
	{
		lead_update_poly();
	}
	else
	{
		lead_update_mono();
	}
}