//SuBoard by shiru8bit (shiru@mail.ru) CC-BY

#include "neslib.h"
#include "apu.h"

#pragma bssseg (push,"ZEROPAGE")
#pragma dataseg(push,"ZEROPAGE")

unsigned char oam_off;
unsigned char upd_off;
unsigned char pad;
unsigned char pad_t;

unsigned char game_bright;
unsigned char game_frame;

unsigned char _i,_j,_n;
unsigned char _sx,_sy;
unsigned char _ptr;

#pragma dataseg(pop);
#pragma bssseg (pop);

#include "keyboard.h"
#include "pitch_table.h"
#include "envelope.h"
#include "envelopes.h"
#include "echo.h"
#include "lead.h"
#include "field_nam.h"
#include "pattern.h"
#include "patterns.h"


unsigned char update_list[256];


enum {
	NUMPAD_MODE_OCTAVE=0,
	NUMPAD_MODE_ARP_SPEED,
	NUMPAD_MODE_MONO,
	NUMPAD_MODE_ECHO_DEPTH,
	NUMPAD_MODE_ECHO_LEN,
	NUMPAD_MODE_VOLENV,
	NUMPAD_MODE_DUAL,
	NUMPAD_MODE_PITENV,
	NUMPAD_MODE_PITWHEEL,
	NUMPAD_MODE_PATTERN_NUM,
	NUMPAD_MODE_PATTERN_SPEED,
	NUMPAD_MODE_LEAD_GAIN
};

unsigned char numpad_mode;
const char* numpad_mode_str;
unsigned char numpad_value[2];

const unsigned char palette_bg[16]={ 0x0f,0x30,0x10,0x0c,0x0f,0x00,0x10,0x27,0x0f,0x00,0x10,0x30,0x0f,0x00,0x10,0x30 };

const char numpad_mode_str_octave    []="OCTAVE";
const char numpad_mode_str_arp_speed []="ARPSPD";
const char numpad_mode_str_mono      []="MONOMD";
const char numpad_mode_str_echo_depth[]="ECHODP";
const char numpad_mode_str_echo_len  []="ECHOLN";
const char numpad_mode_str_volenv    []="VOLENV";
const char numpad_mode_str_dual      []="DUALMD";
const char numpad_mode_str_pitenv    []="PITENV";
const char numpad_mode_str_pitwhl    []="PITWHL";
const char numpad_mode_str_ptnnum    []="PTNNUM";
const char numpad_mode_str_ptnspd    []="PTNSPD";
const char numpad_mode_str_lead_gain []="LDGAIN";

unsigned char keyb_raw_copy[KBD_ROWS*2];
unsigned char keyb_fade;
unsigned char keyb_color;
unsigned char keyb_any;



unsigned char volenv_id;
unsigned char volenv_id_prev;
	
const signed char const* env_volume_list[]={
	env_volume_1,
	env_volume_2,
	env_volume_3,
	env_volume_4,
	env_volume_5,
	env_volume_6,
	env_volume_7,
	env_volume_8,
	env_volume_9,
};
	
unsigned char pitenv_id;
unsigned char pitenv_id_prev;
	
const signed char const* env_pitch_list[]={
	env_pitch_1,
	env_pitch_2,
	env_pitch_3,
	env_pitch_4,
	env_pitch_5,
	env_pitch_6,
	env_pitch_7,
	env_pitch_8,
	env_pitch_9,
};

unsigned char pattern_id;
unsigned char pattern_id_prev;

const unsigned int* const patterns_list[]={
	pattern_1,
	pattern_2,
	pattern_3,
	pattern_4,
	pattern_5,
	pattern_6,
	pattern_7,
	pattern_8,
	pattern_9
};

unsigned char trigger_keys[16];



unsigned char trigger_set(unsigned char id)
{
	if(trigger_keys[id]==0)
	{
		trigger_keys[id]=1;
		return TRUE;
	}
	
	return FALSE;
}



void trigger_res(unsigned char id)
{
	trigger_keys[id]=0;
}



void put_str(unsigned int adr,const char *str)
{
	vram_adr(adr);

	while(1)
	{
		if(!*str) break;

		vram_put((*str++)-0x20);//-0x20 because ASCII code 0x20 is placed in tile 0 of the CHR
	}
}



void main(void)
{
	kb_init();
	apu_init();
	
	oam_clear();
	
	pal_bg(palette_bg);
	pal_spr(palette_bg);

	vram_adr(NAMETABLE_A);
	vram_unrle(field_nam);
	
	put_str(NTADR_A(2,2),"SUBOARD V1.11");

	game_frame=0;

	numpad_mode=NUMPAD_MODE_OCTAVE;
	numpad_mode_str=numpad_mode_str_octave;
	numpad_value[0]='-';
	numpad_value[1]='-';
	
	memfill(keyb_raw_copy,0,sizeof(keyb_raw_copy));
	keyb_fade=0;
	keyb_color=0x20;
	keyb_any=FALSE;
	
	echo_init();
	lead_init(ppu_system());
	
	pattern_id=0;
	pattern_id_prev=0;
	
	pattern_init(ppu_system());
	pattern_change(patterns_list[pattern_id]);
	
	volenv_id=0;
	volenv_id_prev=0xff;
	pitenv_id=0;
	pitenv_id_prev=0xff;

	memfill(trigger_keys,0,sizeof(trigger_keys));
	
	scroll(-2,0);
	
	ppu_on_bg();
	
	while(1)
	{
		if(volenv_id_prev!=volenv_id)
		{
			lead_volenv_set(env_volume_list[volenv_id]);
			volenv_id_prev=volenv_id;
		}
		
		if(pitenv_id_prev!=pitenv_id)
		{
			lead_pitenv_set(env_pitch_list[pitenv_id]);
			pitenv_id_prev=pitenv_id;
		}
				
		if(!kb_is_any())
		{
			keyb_any=FALSE;
			
			if(keyb_fade<16)
			{
				if(!(keyb_fade&7))
				{
					if(keyb_color>=0x10) keyb_color-=0x10; else keyb_color=palette_bg[3];
				}
				
				++keyb_fade;
				
				if(keyb_fade>=16)
				{
					memfill(keyb_raw_copy,0,sizeof(keyb_raw_copy));
				}
			}
		}
		else
		{
			memcpy(keyb_raw_copy,keyb_raw,sizeof(keyb_raw_copy));
			keyb_fade=0;
			if(!keyb_any)
			{
				keyb_color=0x20+rand8()%14;
				keyb_any=TRUE;
			}
		}
		
		pal_col(7,keyb_color);
		
		upd_off=0;

		if(!(game_frame&1))
		{
			update_list[upd_off++]=MSB(NTADR_A(21,2))|NT_UPD_HORZ;
			update_list[upd_off++]=LSB(NTADR_A(21,2));
			update_list[upd_off++]=6+1+2;
			
			if(!numpad_mode_str)
			{
				for(_i=0;_i<6+1+2;++_i) update_list[upd_off++]=0;
			}
			else
			{
				for(_i=0;_i<6;++_i) update_list[upd_off++]=numpad_mode_str[_i]-32;
				
				update_list[upd_off++]=0;
				update_list[upd_off++]=numpad_value[0]-32;
				update_list[upd_off++]=numpad_value[1]-32;
			}
		}
		else
		{
			update_list[upd_off++]=MSB(NAMETABLE_A+960+8)|NT_UPD_HORZ;
			update_list[upd_off++]=LSB(NAMETABLE_A+960+8);
			update_list[upd_off++]=6*8;
			
			_i=0;
			
			while(_i<KBD_ROWS*2)
			{
				_n=0;
				
				if(keyb_raw_copy[_i]&(1<<4)) _n|=0x05;
				if(keyb_raw_copy[_i]&(1<<3)) _n|=0x50;
				
				update_list[upd_off++]=_n;
				
				_n=0;
				
				if(keyb_raw_copy[_i]&(1<<2)) _n|=0x05;
				if(keyb_raw_copy[_i]&(1<<1)) _n|=0x50;
				
				update_list[upd_off++]=_n;
				
				++_i;
				
				if(_i==9||_i==18) ++_i;
			}
		}
		
		update_list[upd_off++]=NT_UPD_EOF;
		
		set_vram_update(update_list);
		
		//update in the same frame, should improve responsivity a bit, but add some jitter

		apu_clear();
		pattern_update();
		lead_update();
		echo_update();
		apu_update();
		
		ppu_wait_nmi();
		
		pad_t=pad_trigger(0)|pad_trigger(1);
		pad=pad_state(0)|pad_state(1);
		
		kb_poll();
		
		arp_list_clear();

		if(kb_state(KB_INS))       numpad_mode=NUMPAD_MODE_OCTAVE;
		if(kb_state(KB_HOME))      numpad_mode=NUMPAD_MODE_ARP_SPEED;
		if(kb_state(KB_DEL))       numpad_mode=NUMPAD_MODE_MONO;
		if(kb_state(KB_PGUP))      numpad_mode=NUMPAD_MODE_ECHO_DEPTH;
		if(kb_state(KB_PGDOWN))    numpad_mode=NUMPAD_MODE_ECHO_LEN;
		if(kb_state(KB_NUMLOCK))   numpad_mode=NUMPAD_MODE_VOLENV;
		if(kb_state(KB_PAD_PLUS))  numpad_mode=NUMPAD_MODE_DUAL;
		if(kb_state(KB_PAD_MINUS)) numpad_mode=NUMPAD_MODE_PITENV;
		if(kb_state(KB_END))       numpad_mode=NUMPAD_MODE_PITWHEEL;
		if(kb_state(KB_PAD_0))     numpad_mode=NUMPAD_MODE_PATTERN_NUM;
		if(kb_state(KB_PAD_DEL))   numpad_mode=NUMPAD_MODE_PATTERN_SPEED;
		if(kb_state(KB_BACKSPACE)) numpad_mode=NUMPAD_MODE_LEAD_GAIN;
		
		if(kb_state(KB_Z)) arp_list_add(NOTE_C);
		if(kb_state(KB_S)) arp_list_add(NOTE_Ch);
		if(kb_state(KB_X)) arp_list_add(NOTE_D);
		if(kb_state(KB_D)) arp_list_add(NOTE_Dh);
		if(kb_state(KB_C)) arp_list_add(NOTE_E);
		if(kb_state(KB_V)) arp_list_add(NOTE_F);
		if(kb_state(KB_G)) arp_list_add(NOTE_Fh);
		if(kb_state(KB_B)) arp_list_add(NOTE_G);
		if(kb_state(KB_H)) arp_list_add(NOTE_Gh);
		if(kb_state(KB_N)) arp_list_add(NOTE_A);
		if(kb_state(KB_J)) arp_list_add(NOTE_Ah);
		if(kb_state(KB_M)) arp_list_add(NOTE_B);
		if(kb_state(KB_COMMA)) arp_list_add(12+NOTE_C);
		if(kb_state(KB_L))     arp_list_add(12+NOTE_Ch);
		if(kb_state(KB_DOT))   arp_list_add(12+NOTE_D);
		if(kb_state(KB_COLON)) arp_list_add(12+NOTE_Dh);
		if(kb_state(KB_SLASH)) arp_list_add(12+NOTE_E);
			
		if(kb_state(KB_Q)) arp_list_add(12+NOTE_C);
		if(kb_state(KB_2)) arp_list_add(12+NOTE_Ch);
		if(kb_state(KB_W)) arp_list_add(12+NOTE_D);
		if(kb_state(KB_3)) arp_list_add(12+NOTE_Dh);
		if(kb_state(KB_E)) arp_list_add(12+NOTE_E);
		if(kb_state(KB_R)) arp_list_add(12+NOTE_F);
		if(kb_state(KB_5)) arp_list_add(12+NOTE_Fh);
		if(kb_state(KB_T)) arp_list_add(12+NOTE_G);
		if(kb_state(KB_6)) arp_list_add(12+NOTE_Gh);
		if(kb_state(KB_Y)) arp_list_add(12+NOTE_A);
		if(kb_state(KB_7)) arp_list_add(12+NOTE_Ah);
		if(kb_state(KB_U)) arp_list_add(12+NOTE_B);
		if(kb_state(KB_I)) arp_list_add(24+NOTE_C);
		if(kb_state(KB_9)) arp_list_add(24+NOTE_Ch);
		if(kb_state(KB_O)) arp_list_add(24+NOTE_D);
		if(kb_state(KB_0)) arp_list_add(24+NOTE_Dh);
		if(kb_state(KB_P)) arp_list_add(24+NOTE_E);
		if(kb_state(KB_OPENBRACE))  arp_list_add(24+NOTE_F);
		if(kb_state(KB_EQUALS))     arp_list_add(24+NOTE_Fh);
		if(kb_state(KB_CLOSEBRACE)) arp_list_add(24+NOTE_G);
		if(kb_state(KB_BACKSLASH))  arp_list_add(24+NOTE_Gh);
		
		if(kb_state(KB_CTRL)||kb_state(KB_SHIFT)||(pad&(PAD_A|PAD_B)))	//change bass note
		{
			if(arp_list_len>0)
			{
				pattern_bass_base=arp_list[0];
			
				if(kb_state(KB_SHIFT)||(pad&PAD_B)) arp_list_len=0;	//mute lead
			}
		}
		
		pitch_wheel_to=0;

		if(kb_state(KB_UP)   ||(pad&PAD_UP))    pitch_wheel_to=16*2;
		if(kb_state(KB_LEFT) ||(pad&PAD_LEFT))  pitch_wheel_to=16;
		if(kb_state(KB_RIGHT)||(pad&PAD_RIGHT)) pitch_wheel_to=-16;
		if(kb_state(KB_DOWN) ||(pad&PAD_DOWN))  pitch_wheel_to=-16*2;

		if(kb_state(KB_PAD_SLASH)) lead_mute_at_release=TRUE;
		if(kb_state(KB_PAD_ASTERISK)) lead_mute_at_release=FALSE;
		
		if(kb_state(KB_SPACE)) lead_env_stop(TRUE);
		
		if(kb_state(KB_F5)||(pad_t&PAD_START)) pattern_start();
		if(kb_state(KB_F6)||(pad_t&PAD_SELECT)) pattern_stop();
		
		if(kb_state(KB_F7)) { if(trigger_set(0)) pattern_noise_period=(pattern_noise_period&0xf0)|((pattern_noise_period+1)&0x0f); } else trigger_res(0);
		if(kb_state(KB_F8)) { if(trigger_set(1)) pattern_noise_period=(pattern_noise_period&0xf0)|((pattern_noise_period+15)&0x0f); } else trigger_res(1);
		if(kb_state(KB_F9)) { if(trigger_set(2)) pattern_noise_period^=0x80; } else trigger_res(2);
		
		if(kb_state(KB_F10)) { if(trigger_set(3)) pattern_voice^=PATTERN_CH_DMC;   } else trigger_res(3);
		if(kb_state(KB_F11)) { if(trigger_set(4)) pattern_voice^=PATTERN_CH_NOISE; } else trigger_res(4);
		if(kb_state(KB_F12)) { if(trigger_set(5)) pattern_voice^=PATTERN_CH_BASS;  } else trigger_res(5);
		
		if(kb_state(KB_RETURN)) pattern_break();
		
		switch(numpad_mode)
		{
		case NUMPAD_MODE_OCTAVE:
			{
				if(kb_state(KB_PAD_1)) base_octave=0;
				if(kb_state(KB_PAD_2)) base_octave=1;
				if(kb_state(KB_PAD_3)) base_octave=2;
				if(kb_state(KB_PAD_4)) base_octave=3;
				if(kb_state(KB_PAD_5)) base_octave=4;
				if(kb_state(KB_PAD_6)) base_octave=5;
				
				numpad_mode_str=numpad_mode_str_octave;
				numpad_value[0]='0';
				numpad_value[1]='1'+base_octave;
			}
			break;
			
		case NUMPAD_MODE_ARP_SPEED:
			{
				if(kb_state(KB_PAD_1)) arp_speed=1;
				if(kb_state(KB_PAD_2)) arp_speed=2;
				if(kb_state(KB_PAD_3)) arp_speed=3;
				if(kb_state(KB_PAD_4)) arp_speed=4;
				if(kb_state(KB_PAD_5)) arp_speed=5;
				if(kb_state(KB_PAD_6)) arp_speed=6;
				
				numpad_mode_str=numpad_mode_str_arp_speed;
				numpad_value[0]='0';
				numpad_value[1]='0'+arp_speed;
			}
			break;
			
		case NUMPAD_MODE_MONO:
			{
				if(kb_state(KB_PAD_1)) mono_mode=MONOMODE_ARP_MONO;
				if(kb_state(KB_PAD_2)) mono_mode=MONOMODE_ARP_LEGATO; 
				if(kb_state(KB_PAD_3)) mono_mode=MONOMODE_MONO; 
				if(kb_state(KB_PAD_4)) { mono_mode=MONOMODE_PORTA; porta_speed=64; }
				if(kb_state(KB_PAD_5)) { mono_mode=MONOMODE_PORTA; porta_speed=32; }
				if(kb_state(KB_PAD_6)) { mono_mode=MONOMODE_PORTA; porta_speed=16; }
				if(kb_state(KB_PAD_7)) { mono_mode=MONOMODE_PORTA; porta_speed=8;  }
				if(kb_state(KB_PAD_8)) { mono_mode=MONOMODE_PORTA; porta_speed=4; }
				if(kb_state(KB_PAD_9)) mono_mode=MONOMODE_POLY;
				
				numpad_mode_str=numpad_mode_str_mono;
				
				switch(mono_mode)
				{
				case MONOMODE_ARP_MONO:  	numpad_value[0]='A'; numpad_value[1]='M'; break;
				case MONOMODE_ARP_LEGATO: 	numpad_value[0]='A'; numpad_value[1]='L'; break;
				case MONOMODE_MONO: 		numpad_value[0]='M'; numpad_value[1]='N'; break;
				case MONOMODE_PORTA:
					switch(porta_speed)
					{
					case 64:   numpad_value[0]='P'; numpad_value[1]='5'; break;
					case 32:   numpad_value[0]='P'; numpad_value[1]='6'; break;
					case 16:   numpad_value[0]='P'; numpad_value[1]='7'; break;
					case 8:    numpad_value[0]='P'; numpad_value[1]='8'; break;
					case 4:    numpad_value[0]='P'; numpad_value[1]='9'; break;
					}
					break;
				case MONOMODE_POLY: numpad_value[0]='P'; numpad_value[1]='L'; break;
				}
			}
			break;
			
		case NUMPAD_MODE_ECHO_DEPTH:
			{
				if(kb_state(KB_PAD_1)) echo_fade=16;
				if(kb_state(KB_PAD_2)) echo_fade=6;
				if(kb_state(KB_PAD_3)) echo_fade=5;
				if(kb_state(KB_PAD_4)) echo_fade=4;
				if(kb_state(KB_PAD_5)) echo_fade=3;
				if(kb_state(KB_PAD_6)) echo_fade=2;
				if(kb_state(KB_PAD_7)) echo_fade=1;
				
				numpad_mode_str=numpad_mode_str_echo_depth;
				
				switch(echo_fade)
				{
				case 16: 
					numpad_value[0]='0'; numpad_value[1]='0';
					break;
				default:
					numpad_value[0]='0';
					numpad_value[1]='0'+(7-echo_fade);
				}
			}
			break;
			
		case NUMPAD_MODE_ECHO_LEN:
			{
				if(kb_state(KB_PAD_1)) echo_len=32;
				if(kb_state(KB_PAD_2)) echo_len=24;
				if(kb_state(KB_PAD_3)) echo_len=16;
				if(kb_state(KB_PAD_4)) echo_len=14;
				if(kb_state(KB_PAD_5)) echo_len=12;
				if(kb_state(KB_PAD_6)) echo_len=10;
				if(kb_state(KB_PAD_7)) echo_len=8;
				if(kb_state(KB_PAD_8)) echo_len=6;
				if(kb_state(KB_PAD_9)) echo_len=4;
				
				numpad_mode_str=numpad_mode_str_echo_len;
				numpad_value[0]=echo_len/10%10+'0';
				numpad_value[1]=echo_len%10+'0';
			}
			break;
			
		case NUMPAD_MODE_VOLENV:
			{
				if(kb_state(KB_PAD_1)) volenv_id=0;
				if(kb_state(KB_PAD_2)) volenv_id=1;
				if(kb_state(KB_PAD_3)) volenv_id=2;
				if(kb_state(KB_PAD_4)) volenv_id=3;
				if(kb_state(KB_PAD_5)) volenv_id=4;
				if(kb_state(KB_PAD_6)) volenv_id=5;
				if(kb_state(KB_PAD_7)) volenv_id=6;
				if(kb_state(KB_PAD_8)) volenv_id=7;
				if(kb_state(KB_PAD_9)) volenv_id=8;
				
				numpad_mode_str=numpad_mode_str_volenv;
				numpad_value[0]='0';
				numpad_value[1]='1'+volenv_id;
			}
			break;
			
		case NUMPAD_MODE_PITENV:
			{
				if(kb_state(KB_PAD_1)) pitenv_id=0;
				if(kb_state(KB_PAD_2)) pitenv_id=1;
				if(kb_state(KB_PAD_3)) pitenv_id=2;
				if(kb_state(KB_PAD_4)) pitenv_id=3;
				if(kb_state(KB_PAD_5)) pitenv_id=4;
				if(kb_state(KB_PAD_6)) pitenv_id=5;
				if(kb_state(KB_PAD_7)) pitenv_id=6;
				if(kb_state(KB_PAD_8)) pitenv_id=7;
				if(kb_state(KB_PAD_9)) pitenv_id=8;
				
				numpad_mode_str=numpad_mode_str_pitenv;
				numpad_value[0]='0';
				numpad_value[1]='1'+pitenv_id;
			}
			break;
			
		case NUMPAD_MODE_PITWHEEL:
			{
				if(kb_state(KB_PAD_1)) pitch_wheel_speed=32;
				if(kb_state(KB_PAD_2)) pitch_wheel_speed=16;
				if(kb_state(KB_PAD_3)) pitch_wheel_speed=8;
				if(kb_state(KB_PAD_4)) pitch_wheel_speed=7;
				if(kb_state(KB_PAD_5)) pitch_wheel_speed=6;
				if(kb_state(KB_PAD_6)) pitch_wheel_speed=5;
				if(kb_state(KB_PAD_7)) pitch_wheel_speed=4;
				if(kb_state(KB_PAD_8)) pitch_wheel_speed=3;
				if(kb_state(KB_PAD_9)) pitch_wheel_speed=2;
				
				numpad_mode_str=numpad_mode_str_pitwhl;
				numpad_value[0]='0'+pitch_wheel_speed/100;
				numpad_value[1]='1'+pitch_wheel_speed%10;
			}
			break;
			
		case NUMPAD_MODE_PATTERN_NUM:
			{
				if(kb_state(KB_PAD_1)) pattern_id=0;
				if(kb_state(KB_PAD_2)) pattern_id=1;
				if(kb_state(KB_PAD_3)) pattern_id=2;
				if(kb_state(KB_PAD_4)) pattern_id=3;
				if(kb_state(KB_PAD_5)) pattern_id=4;
				if(kb_state(KB_PAD_6)) pattern_id=5;
				if(kb_state(KB_PAD_7)) pattern_id=6;
				if(kb_state(KB_PAD_8)) pattern_id=7;
				if(kb_state(KB_PAD_9)) pattern_id=8;
				
				if(pattern_id_prev!=pattern_id)
				{
					pattern_change(patterns_list[pattern_id]);
					pattern_id_prev=pattern_id;
				}
				
				numpad_mode_str=numpad_mode_str_ptnnum;
				numpad_value[0]='0';
				numpad_value[1]='1'+pattern_id;
			}
			break;
			
		case NUMPAD_MODE_PATTERN_SPEED:
			{
				if(kb_state(KB_PAD_1)) pattern_tempo=1;
				if(kb_state(KB_PAD_2)) pattern_tempo=2;
				if(kb_state(KB_PAD_3)) pattern_tempo=3;
				if(kb_state(KB_PAD_4)) pattern_tempo=4;
				if(kb_state(KB_PAD_5)) pattern_tempo=5;
				if(kb_state(KB_PAD_6)) pattern_tempo=6;
				if(kb_state(KB_PAD_7)) pattern_tempo=7;
				if(kb_state(KB_PAD_8)) pattern_tempo=8;
				if(kb_state(KB_PAD_9)) pattern_tempo=9;
				
				numpad_mode_str=numpad_mode_str_ptnspd;
				numpad_value[0]='0';
				numpad_value[1]='0'+pattern_tempo;
			}
			break;
			
		case NUMPAD_MODE_DUAL:
			{
				if(kb_state(KB_PAD_1)) lead_dual_offset=0;
				if(kb_state(KB_PAD_2)) lead_dual_offset=1;
				if(kb_state(KB_PAD_3)) lead_dual_offset=2;
				if(kb_state(KB_PAD_4)) lead_dual_offset=4;
				if(kb_state(KB_PAD_5)) lead_dual_offset=3*16;
				if(kb_state(KB_PAD_6)) lead_dual_offset=5*16;
				if(kb_state(KB_PAD_7)) lead_dual_offset=7*16;
				if(kb_state(KB_PAD_8)) lead_dual_offset=12*16;
				if(kb_state(KB_PAD_9)) lead_dual_offset=24*16;
				
				numpad_mode_str=numpad_mode_str_dual;
				numpad_value[0]='0';
				
				switch(lead_dual_offset)
				{
				case 0: numpad_value[1]='0'; break;
				case 1: numpad_value[1]='1'; break;
				case 2: numpad_value[1]='2'; break;
				case 4: numpad_value[1]='3'; break;
				case 3*16: numpad_value[1]='4'; break;
				case 5*16: numpad_value[1]='5'; break;
				case 7*16: numpad_value[1]='6'; break;
				case 12*16: numpad_value[1]='7'; break;
				case 12*16-2: numpad_value[1]='8'; break;
				}
			}
			break;
			
		case NUMPAD_MODE_LEAD_GAIN:
			{
				if(kb_state(KB_PAD_1)) lead_gain=4;
				if(kb_state(KB_PAD_2)) lead_gain=5;
				if(kb_state(KB_PAD_3)) lead_gain=6;
				if(kb_state(KB_PAD_4)) lead_gain=7;
				if(kb_state(KB_PAD_5)) lead_gain=8;
				if(kb_state(KB_PAD_6)) lead_gain=9;
				if(kb_state(KB_PAD_7)) lead_gain=11;
				if(kb_state(KB_PAD_8)) lead_gain=13;
				if(kb_state(KB_PAD_9)) lead_gain=15;
				
				numpad_mode_str=numpad_mode_str_lead_gain;
				numpad_value[0]='0';
				
				switch(lead_gain)
				{
				case 4: numpad_value[1]='1'; break;
				case 5: numpad_value[1]='2'; break;
				case 6: numpad_value[1]='3'; break;
				case 7: numpad_value[1]='4'; break;
				case 8: numpad_value[1]='5'; break;
				case 9: numpad_value[1]='6'; break;
				case 11: numpad_value[1]='7'; break;
				case 13: numpad_value[1]='8'; break;
				case 15: numpad_value[1]='9'; break;
				}
			}
			break;
		}
		
		if(kb_state(KB_F1)) lead_duty=2<<6;
		if(kb_state(KB_F2)) lead_duty=1<<6;
		if(kb_state(KB_F3)) lead_duty=0<<6;
		if(kb_state(KB_F4)) lead_duty=0xff;	//triangle mode

		++game_frame;
	}
}