extern const unsigned char* s_kick_adr;
extern const unsigned char* s_snare_adr;
extern const unsigned char* s_rim_adr;
extern const unsigned char* s_clap_adr;
extern const unsigned char* s_tom_adr;
extern const unsigned char* s_cowbell_adr;

extern unsigned char s_kick_len;
extern unsigned char s_snare_len;
extern unsigned char s_rim_len;
extern unsigned char s_clap_len;
extern unsigned char s_tom_len;
extern unsigned char s_cowbell_len;


#define PATTERN_LEN			16

#define PATTERN_CH_DMC		0x01
#define PATTERN_CH_NOISE	0x02
#define PATTERN_CH_BASS		0x04

#define BD			0x0001
#define SD			0x0002
#define RM			0x0004
#define CL			0x0008
#define TM			0x0010
#define CW			0x0020
#define CH			0x0040
#define OH			0x0080
#define BASS(x)		(((x)+1)<<8)
#define BASS_REST	0xff00



unsigned char pattern_voice;

unsigned char pattern_tempo;
unsigned char pattern_tempo_cnt;

const unsigned int* pattern_data;
const unsigned int* pattern_data_next;

unsigned char pattern_row;
unsigned char pattern_row_off;

unsigned char pattern_noise_period;
unsigned char pattern_active;

const unsigned int* pattern_pitch_table;
unsigned int pattern_bass_period;
unsigned int pattern_bass_base;

struct env_struct env_noise_vol;


	
void dmc_start(const unsigned char* dmc_data,unsigned char dmc_len,unsigned char dmc_pitch)
{
	APU_SND_CHN(0x0f);		//stop DMC

	APU_DMC_START(((unsigned int)dmc_data)>>6);
	APU_DMC_LEN  (dmc_len);
	APU_DMC_FREQ (dmc_pitch);
	
	APU_SND_CHN(0x1f);		//restart DMC
}



void pattern_init(unsigned char ntsc)
{
	pattern_voice=PATTERN_CH_DMC|PATTERN_CH_NOISE|PATTERN_CH_BASS;
	pattern_tempo=6;
	pattern_tempo_cnt=0;
	
	pattern_pitch_table=(ntsc?pitch_table_ntsc:pitch_table_pal);
	
	pattern_data=NULL;
	pattern_data_next=NULL;
	pattern_row=0;
	pattern_row_off=0;
	pattern_active=FALSE;
	pattern_bass_period=0;
	pattern_bass_base=12;
	
	pattern_noise_period=0;
	
	env_set_data(&env_noise_vol,0);
	env_stop(&env_noise_vol);
}



void pattern_change(const unsigned int* data)
{
	pattern_data_next=data;
	
	if(!pattern_active) pattern_data=pattern_data_next;
}



void pattern_break(void)
{
	pattern_row_off=PATTERN_LEN;
}



void pattern_start(void)
{
	if(pattern_active) return;
	
	pattern_row=0;
	pattern_row_off=0;
	pattern_tempo_cnt=0;
	pattern_active=TRUE;
}



void pattern_stop(void)
{
	if(!pattern_active) return;
	
	pattern_bass_period=0;
	pattern_active=FALSE;
}



void pattern_update(void)
{
	unsigned char bass1,bass2;
	unsigned int row;
	
	env_update(&env_noise_vol);
	
	if(pattern_active&&pattern_data)
	{
		if(!pattern_tempo_cnt)
		{
			if(pattern_row>=PATTERN_LEN)
			{
				pattern_data=pattern_data_next;
				pattern_row=0;
				pattern_row_off=0;
			}
					
			row=pattern_data[pattern_row+pattern_row_off];
			
			if(pattern_voice&PATTERN_CH_DMC)
			{
				if(row&BD) dmc_start(s_kick_adr   ,s_kick_len   ,0x0f);
				if(row&SD) dmc_start(s_snare_adr  ,s_snare_len  ,0x0f);
				if(row&RM) dmc_start(s_rim_adr    ,s_rim_len    ,0x0f);
				if(row&CL) dmc_start(s_clap_adr   ,s_clap_len   ,0x0f);
				if(row&TM) dmc_start(s_tom_adr    ,s_tom_len    ,0x0f);
				if(row&CW) dmc_start(s_cowbell_adr,s_cowbell_len,0x0f);
			}
			
			if(row&CH)
			{
				env_set_data(&env_noise_vol,env_noise_volume_CH);
				env_start(&env_noise_vol);
			}
			
			if(row&OH)
			{
				env_set_data(&env_noise_vol,env_noise_volume_OH);
				env_start(&env_noise_vol);
			}
			
			row>>=8;
			
			if(row)
			{
				if(row==0xff)	//stop bass note
				{
					pattern_bass_period=0;
				}
				else	//start bass note
				{
					pattern_bass_period=pattern_pitch_table[(pattern_bass_base+(row-1))*16];
				}
			}
		}
		
		if(pattern_tempo_cnt==(pattern_tempo-1))
		{
			bass1=pattern_data[pattern_row+pattern_row_off]>>8;
			bass2=pattern_data[((pattern_row+1)%PATTERN_LEN)+pattern_row_off]>>8;
			
			if(bass2&&(bass2!=0xff)&&(bass1==bass2))
			{
				pattern_bass_period=0;	//a frame gap between same repeating notes
			}
			
			++pattern_row;
		}
	}
	
	if((pattern_voice&PATTERN_CH_NOISE))
	{
		noise_vol=env_noise_vol.out;
		noise_period=pattern_noise_period;	
	}
	else
	{
		noise_vol=0;
	}
	
	if(pattern_bass_period&&(pattern_voice&PATTERN_CH_BASS))
	{
		tri_vol=15;
		tri_period=pattern_bass_period;
	}
	else
	{
		tri_vol=0;
	}
	
	++pattern_tempo_cnt;
	
	if(pattern_tempo_cnt>=pattern_tempo) pattern_tempo_cnt=0;
}